
#include <bits/stdc++.h>

using namespace std;

int32_t main() {
  ios::sync_with_stdio(false);
  cin.tie(0);

  int n;
  cin >> n;

  for (int i = 1; i <= 100000; i++) { // more than output limit of 100kb
    cout << 123456 << "\n";
  }
  return 0;
}
